from collections import deque
from functools import lru_cache
from dcicutils.function_cache_decorator import function_cache
import io
import json
from pyramid.config import Configurator as PyramidConfigurator
from pyramid.paster import get_app as pyramid_get_app
from pyramid.response import Response as PyramidResponse
from pyramid.router import Router as PyramidRouter
import os
import re
import requests
from requests.models import Response
from threading import Thread
from typing import Callable, Dict, List, Optional, Tuple, Type, Union
# from urllib.parse import parse_qs as parse_url_query_string
from uuid import uuid4 as uuid
from webtest.app import TestApp, TestResponse
from wsgiref.simple_server import make_server as wsgi_make_server
from dcicutils.common import APP_SMAHT, OrchestratedApp, ORCHESTRATED_APPS
from dcicutils.ff_utils import get_metadata, get_schema, patch_metadata, post_metadata
from dcicutils.misc_utils import to_camel_case, VirtualApp
from dcicutils.schema_utils import get_identifying_properties
from dcicutils.tmpfile_utils import temporary_file

Portal = Type["Portal"]  # Forward type reference for type hints.
OptionalResponse = Optional[Union[Response, TestResponse]]


class Portal:
    """
    This is meant to be an Über wrapper for Portal access. It can be created in a variety of ways:
    1. From a (Portal) .ini file (e.g. development.ini).
    2. From a key dictionary, containing "key" and "secret" and (optional) "server" property values.
    3. From a key pair tuple, containing (in order) a key and secret values.
    4. From a keys .json file residing in ~/.{app}-keys.json where the given "app" value is either "smaht", "cgap",
       or "fourfront"; where is assumed to contain a dictionary with a key for the given "env" value, e.g. smaht-local;
       and with a dictionary value containing "key" and "secret" property values, and an optional "server" property;
       if an "app" value is not specified but the given "env" value begins with one of the app values then that value
       will be used, i.e. e.g. if "env" is "smaht-local" and app is unspecified than app is assumed to be "smaht".
    5. From a keys .json file as described above (#4) but rather than be identified by the given "env" value it
       is looked up via the given "server" name and the "server" key dictionary value in the key file.
    6. From a full path to a keys .json file.
    7. From a given "vapp" value; which may be a webtest.app.TestApp,
       or a dcicutils.misc_utils.VirtualApp, or even a pyramid.router.Router.
    8. From another Portal object (i.e. copy constructor).
    """
    DEFAULT_APP = APP_SMAHT
    KEYS_FILE_DIRECTORY = "~"
    MIME_TYPE_JSON = "application/json"
    FILE_TYPE_SCHEMA_NAME = "File"

    # Object lookup strategies; on a per-reference (type/value) basis, used currently ONLY by
    # structured_data.py; controlled by an optional lookup_strategy callable; default is
    # lookup at root path but after the specified type path lookup, and then lookup all subtypes;
    # can choose to lookup root path first, or not lookup root path at all, or not lookup
    # subtypes at all; the lookup_strategy callable if specified should take a type_name
    # and value (string) arguements and return an integer of any of the below ORed together.
    # The main purpose of this is optimization; to minimize portal lookups; since for example,
    # currently at least, /{type}/{accession} does not work but /{accession} does; so we
    # currently (smaht-portal/.../ingestion_processors) use LOOKUP_ROOT_FIRST for this.
    # And current usage NEVER has LOOKUP_SUBTYPES turned OFF; but support just in case.
    LOOKUP_UNDEFINED = 0
    LOOKUP_SPECIFIED_TYPE = 0x0001
    LOOKUP_ROOT = 0x0002
    LOOKUP_ROOT_FIRST = 0x0004 | LOOKUP_ROOT
    LOOKUP_SUBTYPES = 0x0008
    LOOKUP_DEFAULT = LOOKUP_SPECIFIED_TYPE | LOOKUP_ROOT | LOOKUP_SUBTYPES

    def __init__(self,
                 arg: Optional[Union[Portal, TestApp, VirtualApp, PyramidRouter, dict, tuple, str]] = None,
                 env: Optional[str] = None, server: Optional[str] = None,
                 app: Optional[OrchestratedApp] = None,
                 raise_exception: bool = True) -> None:

        def init(unspecified: Optional[list] = []) -> None:
            self._ini_file = None
            self._key = None
            self._keys_file = None
            self._env = None
            self._server = None
            self._vapp = None
            for arg in unspecified:
                if arg is not None:
                    raise Exception("Portal initialization error; extraneous arguments.")

        def init_from_portal(portal: Portal, unspecified: Optional[list] = None) -> None:
            init(unspecified=unspecified)
            self._ini_file = portal._ini_file
            self._key = portal._key
            self._keys_file = portal._keys_file
            self._env = portal._env
            self._server = portal._server
            self._vapp = portal._vapp
            self._app = portal._app

        def init_from_vapp(vapp: Union[TestApp, VirtualApp, PyramidRouter], unspecified: Optional[list] = []) -> None:
            init(unspecified=unspecified)
            self._vapp = Portal._create_vapp(vapp)

        def init_from_ini_file(ini_file: str, unspecified: Optional[list] = []) -> None:
            init(unspecified=unspecified)
            self._ini_file = ini_file
            self._vapp = Portal._create_vapp(ini_file)

        def init_from_key(key: dict, server: Optional[str], unspecified: Optional[list] = []) -> None:
            init(unspecified=unspecified)
            if (isinstance(key_id := key.get("key"), str) and key_id and
                isinstance(secret := key.get("secret"), str) and secret):  # noqa
                self._key = {"key": key_id, "secret": secret}
                if (isinstance(server, str) and server) or (isinstance(server := key.get("server"), str) and server):
                    if server := Portal._normalize_server(server):
                        if isinstance(key_server := key.get("server"), str) and key_server:
                            if Portal._normalize_server(key_server) != server:
                                raise Exception(f"Portal server inconsistency: {server} vs {key_server}")
                        self._key["server"] = self._server = server
            if not self._key:
                raise Exception("Portal initialization error; from key.")

        def init_from_key_pair(key_pair: tuple, server: Optional[str], unspecified: Optional[list] = []) -> None:
            if len(key_pair) >= 2:
                init_from_key({"key": key_pair[0], "secret": key_pair[1]}, server, unspecified=unspecified)
            else:
                raise Exception("Portal initialization error; from key-pair.")

        def init_from_keys_file(keys_file: str, env: Optional[str], server: Optional[str],
                                unspecified: Optional[list] = []) -> None:
            key, env = Portal._lookup_in_keys_file(keys_file, env, server, raise_exception=True)
            if key:
                init_from_key(key, server)
                self._keys_file = keys_file
                self._env = env

        def init_from_env_server_app(env: str, server: str, app: Optional[str],
                                     unspecified: Optional[list] = None) -> None:
            if keys_file := Portal._default_keys_file(app, env, server):
                init_from_keys_file(keys_file, env, server, unspecified=unspecified)
            else:
                init(unspecified=unspecified)
                self._env = env
                self._server = server

        if (valid_app := app) and not (valid_app := Portal._valid_app(app)):
            raise Exception(f"Portal initialization error; invalid app: {app}")
        self._app = valid_app
        if isinstance(arg, Portal):
            init_from_portal(arg, unspecified=[env, server, app])
        elif isinstance(arg, (TestApp, VirtualApp, PyramidRouter)):
            init_from_vapp(arg, unspecified=[env, server, app])
        elif isinstance(arg, str) and arg.endswith(".ini"):
            init_from_ini_file(arg, unspecified=[env, server])
        elif isinstance(arg, dict):
            init_from_key(arg, server, unspecified=[env])
        elif isinstance(arg, tuple):
            init_from_key_pair(arg, server, unspecified=[env])
        elif isinstance(arg, str) and arg.endswith(".json"):
            init_from_keys_file(arg, env, server)
        elif isinstance(arg, str) and arg:
            init_from_env_server_app(arg, server, app, unspecified=[env])
        elif (isinstance(env, str) and env) or (isinstance(server, str) and server):
            init_from_env_server_app(env, server, app, unspecified=[arg])
        elif not arg and (keys_file := Portal._default_keys_file(self._app or Portal.DEFAULT_APP, env, server)):
            # If no initial arg then look for default app keys file.
            init_from_keys_file(keys_file, env, server)
        elif raise_exception:
            raise Exception("Portal initialization error; insufficient arguments.")
        else:
            init()
        if not self.vapp and not self.key and raise_exception:
            raise Exception("Portal initialization error; neither key nor vapp defined.")

    @property
    def ini_file(self) -> Optional[str]:
        return self._ini_file

    @property
    def key(self) -> Optional[dict]:
        return self._key

    @property
    def key_pair(self) -> Optional[Tuple[str, str]]:
        return (key.get("key"), key.get("secret")) if (key := self.key) else None

    @property
    def key_id(self) -> Optional[str]:
        return key.get("key") if (key := self.key) else None

    @property
    def secret(self) -> Optional[str]:
        return key.get("secret") if (key := self.key) else None

    @property
    def keys_file(self) -> Optional[str]:
        return self._keys_file

    @property
    def env(self) -> Optional[str]:
        return self._env

    @property
    def server(self) -> Optional[str]:
        return key.get("server") if (key := self.key) and key.get("server") else self._server

    @property
    def app(self) -> Optional[str]:
        return self._app

    @property
    def vapp(self) -> Optional[TestApp]:
        return self._vapp

    def get(self, url: str, follow: bool = True,
            raw: bool = False, database: bool = False,
            limit: Optional[int] = None, offset: Optional[int] = None,
            field: Optional[str] = None, deleted: bool = False,
            raise_for_status: bool = False, **kwargs) -> OptionalResponse:
        url = self.url(url, raw, database)
        if isinstance(limit, int) and (limit >= 0):
            if "?" in url:
                url += f"&limit={limit}"
            else:
                url += f"?limit={limit}"
        if isinstance(offset, int) and (offset >= 0):
            if "?" in url:
                url += f"&from={offset}"
            else:
                url += f"?from={offset}"
        if isinstance(field, str) and field:
            if "?" in url:
                url += f"&field={field}"
            else:
                url += f"?field={field}"
        if deleted is True:
            if "?" in url:
                url += "&status=deleted"
            else:
                url += "?status=deleted"
        if not self.vapp:
            response = requests.get(url, allow_redirects=follow, **self._kwargs(**kwargs))
        else:
            response = self.vapp.get(url, **self._kwargs(**kwargs))
            if response and response.status_code in [301, 302, 303, 307, 308] and follow:
                response = response.follow()
            response = self._response(response)
        if raise_for_status:
            response.raise_for_status()
        return response

    def patch(self, url: str, data: Optional[dict] = None, json: Optional[dict] = None,
              raise_for_status: bool = False, **kwargs) -> OptionalResponse:
        url = self.url(url)
        if not self.vapp:
            response = requests.patch(url, data=data, json=json, **self._kwargs(**kwargs))
        else:
            response = self.vapp.patch_json(url, json or data, **self._kwargs(**kwargs))
            response = self._response(response)
        if raise_for_status:
            response.raise_for_status()
        return response

    def post(self, url: str, data: Optional[dict] = None, json: Optional[dict] = None, files: Optional[dict] = None,
             raise_for_status: bool = False, **kwargs) -> OptionalResponse:
        url = self.url(url)
        if files and not ("headers" in kwargs):
            # Setting headers to None when using files implies content-type multipart/form-data.
            kwargs["headers"] = None
        if not self.vapp:
            response = requests.post(url, data=data, json=json, files=files, **self._kwargs(**kwargs))
        else:
            if files:
                response = self.vapp.post(url, json or data, upload_files=files, **self._kwargs(**kwargs))
            else:
                response = self.vapp.post_json(url, json or data, upload_files=files, **self._kwargs(**kwargs))
            response = self._response(response)
        if raise_for_status:
            response.raise_for_status()
        return response

    def get_metadata(self, object_id: str, raw: bool = False, database: bool = False,
                     limit: Optional[int] = None, offset: Optional[int] = None,
                     field: Optional[str] = None, deleted: bool = False,
                     raise_exception: bool = True) -> Optional[dict]:
        if not isinstance(object_id, str):
            return None
        if isinstance(raw, bool) and raw:
            add_on = "frame=raw" + ("&datastore=database" if isinstance(database, bool) and database else "")
        elif database:
            add_on = "datastore=database"
        else:
            add_on = ""
        if isinstance(limit, int) and (limit >= 0):
            if add_on:
                add_on += f"&limit={limit}"
            else:
                add_on += f"limit={limit}"
        if isinstance(offset, int) and (offset >= 0):
            if add_on:
                add_on += f"&from={offset}"
            else:
                add_on += f"from={offset}"
        if isinstance(field, int) and field:
            if add_on:
                add_on += f"&field={field}"
            else:
                add_on += f"field={field}"
        if deleted is True:
            if add_on:
                add_on += "&status=deleted"
            else:
                add_on += "status=deleted"
        if (question_mark := object_id.find("?")) > 0:
            query_string = object_id[question_mark + 1:]
            object_id = object_id[:question_mark]
            if add_on:
                add_on += f"&{query_string}"
            else:
                add_on += query_string
        if raise_exception:
            return get_metadata(obj_id=object_id, vapp=self.vapp, key=self.key, add_on=add_on)
        else:
            try:
                return get_metadata(obj_id=object_id, vapp=self.vapp, key=self.key, add_on=add_on)
            except Exception:
                return None

    def patch_metadata(self, object_id: str, data: dict, check_only: bool = False) -> Optional[dict]:
        if self.key:
            return patch_metadata(obj_id=object_id, patch_item=data, key=self.key,
                                  add_on="check_only=True" if check_only else "")
        return self.patch(f"/{object_id}{'?check_only=True' if check_only else ''}", data).json()

    def post_metadata(self, object_type: str, data: dict, check_only: bool = False) -> Optional[dict]:
        if self.key:
            return post_metadata(schema_name=object_type, post_item=data, key=self.key,
                                 add_on="check_only=True" if check_only else "")
        return self.post(f"/{object_type}{'?check_only=True' if check_only else ''}", data).json()

    def head(self, url: str, follow: bool = True, raise_exception: bool = False, **kwargs) -> Optional[int]:
        try:
            response = requests.head(self.url(url), **self._kwargs(**kwargs))
            if response and response.status_code in [301, 302, 303, 307, 308] and (follow is not False):
                response = response.follow()
            return response.status_code
        except Exception as e:
            if raise_exception is True:
                raise e
        return None

    def get_health(self) -> OptionalResponse:
        return self.get("/health")

    def get_version(self) -> Optional[str]:
        try:
            return self.get_health().json()["project_version"]
        except Exception:
            return None

    def ping(self) -> bool:
        try:
            return self.get_health().status_code == 200
        except Exception:
            return False

    @lru_cache(maxsize=100)
    def get_schema(self, schema_name: str) -> Optional[dict]:
        try:
            return get_schema(self.schema_name(schema_name), portal_vapp=self.vapp, key=self.key)
        except Exception:
            return None

    @lru_cache(maxsize=1)
    def get_schemas(self) -> dict:
        return self.get("/profiles/").json()

    @staticmethod
    @lru_cache(maxsize=100)
    def schema_name(name: str) -> str:
        name = os.path.basename(name).replace(" ", "") if isinstance(name, str) else ""
        if (dot := name.rfind(".")) > 0:
            name = name[0:dot]
        return to_camel_case(name)
        # return to_camel_case(name.replace(" ", "") if not name.endswith(".json") else name[:-5])

    def is_schema_type(self, schema_name_or_portal_object: Union[str, dict], target_schema_name: str,
                       _schemas_super_type_map: Optional[list] = None) -> bool:
        """
        If the given (first) schema_name_or_portal_object argument is a string then returns True iff
        the given schema (type) name isa type of the given target schema (type) name, i.e. the given
        schema type is the given target schema type or has an ancestor which is that type.
        If the given (first) schema_name_or_portal_object argument is a dictionary then
        returns True iff this object value isa type of the given target schema type.
        """
        if isinstance(schema_name_or_portal_object, dict):
            return self.isinstance_schema(schema_name_or_portal_object, target_schema_name)
        elif not isinstance(schema_name_or_portal_object, str) or not schema_name_or_portal_object:
            return False
        schema_name = self.schema_name(schema_name_or_portal_object).lower()
        target_schema_name = self.schema_name(target_schema_name).lower()
        if schema_name == target_schema_name:
            return True
        if super_type_map := (_schemas_super_type_map or self.get_schemas_super_type_map()):
            for super_type in super_type_map:
                if super_type.lower() == target_schema_name:
                    for value in super_type_map[super_type]:
                        if value.lower() == schema_name:
                            return True
        return False

    def is_schema_file_type(self, schema_name_or_portal_object: Union[str, dict]) -> bool:
        return self.is_schema_type(schema_name_or_portal_object, self.FILE_TYPE_SCHEMA_NAME)

    def isinstance_schema(self, portal_object: dict, target_schema_name: str) -> bool:
        """
        Returns True iff the given object isa type of the given schema type.
        """
        if value_types := self.get_schema_types(portal_object):
            schemas_super_type_map = self.get_schemas_super_type_map()
            for value_type in value_types:
                if self.is_schema_type(value_type, target_schema_name, schemas_super_type_map):
                    return True
        return False

    @staticmethod
    def get_schema_types(portal_object: dict) -> Optional[List[str]]:
        if isinstance(portal_object, dict):
            if isinstance(value_types := portal_object.get("@type"), str):
                value_types = [value_types] if value_types else []
            elif not isinstance(value_types, list):
                value_types = []
            if isinstance(data_type := portal_object.get("data_type"), list):
                value_types.extend(data_type)
            elif isinstance(data_type, str):
                value_types.append(data_type)
            return value_types if value_types else None

    @staticmethod
    def get_schema_type(portal_object: dict) -> Optional[str]:
        if value_types := Portal.get_schema_types(portal_object):
            return value_types[0]

    @lru_cache(maxsize=1)
    def get_schemas_super_type_map(self) -> dict:
        """
        Returns the "super type map" for all of the known schemas (via /profiles).
        This is a dictionary with property names which are all known schema type names which
        have (one or more) sub-types, and the value of each such property name is an array
        of all of those sub-type names (direct and all descendents), in breadth first order.
        """
        def list_breadth_first(super_type_map: dict, super_type_name: str) -> dict:
            result = []
            queue = deque(super_type_map.get(super_type_name, []))
            while queue:
                result.append(subtype_name := queue.popleft())
                if subtype_name in super_type_map:
                    queue.extend(super_type_map[subtype_name])
            return result
        if not (schemas := self.get_schemas()):
            return {}
        super_type_map = {}
        for type_name in schemas:
            if isinstance(schema_type := schemas[type_name], dict):
                if super_type_name := schema_type.get("rdfs:subClassOf"):
                    super_type_name = super_type_name.replace("/profiles/", "").replace(".json", "")
                    if super_type_name != "Item":
                        if not super_type_map.get(super_type_name):
                            super_type_map[super_type_name] = [type_name]
                        elif type_name not in super_type_map[super_type_name]:
                            super_type_map[super_type_name].append(type_name)
        super_type_map_flattened = {}
        for super_type_name in super_type_map:
            super_type_map_flattened[super_type_name] = list_breadth_first(super_type_map, super_type_name)
        return super_type_map_flattened

    @lru_cache(maxsize=100)
    def get_schema_subtype_names(self, type_name: str) -> List[str]:
        if not (schemas_super_type_map := self.get_schemas_super_type_map()):
            return []
        return schemas_super_type_map.get(type_name, [])

    @function_cache(maxsize=100, serialize_key=True)
    def get_identifying_paths(self, portal_object: dict, portal_type: Optional[Union[str, dict]] = None,
                              first_only: bool = False,
                              lookup_strategy: Optional[Union[Callable, bool]] = None) -> List[str]:
        """
        Returns the list of the identifying Portal (URL) paths for the given Portal object. Favors any uuid
        and identifier based paths and defavors aliases based paths (ala self.get_identifying_property_names);
        no other ordering defined. Returns an empty list if no identifying properties or otherwise not found.
        Note that this is a newer version of what was in portal_object_utils and just uses the ref_lookup_stratey
        module directly, as it no longer needs to be exposed (to smaht-portal/ingester and smaht-submitr) and so
        this is a first step toward internalizing it to structured_data/portal_utils/portal_object_utils usages.
        """
        def is_lookup_specified_type(lookup_options: int) -> bool:
            return (lookup_options & Portal.LOOKUP_SPECIFIED_TYPE) == Portal.LOOKUP_SPECIFIED_TYPE
        def is_lookup_root(lookup_options: int) -> bool:  # noqa
            return (lookup_options & Portal.LOOKUP_ROOT) == Portal.LOOKUP_ROOT
        def is_lookup_root_first(lookup_options: int) -> bool:  # noqa
            return (lookup_options & Portal.LOOKUP_ROOT_FIRST) == Portal.LOOKUP_ROOT_FIRST
        def is_lookup_subtypes(lookup_options: int) -> bool:  # noqa
            return (lookup_options & Portal.LOOKUP_SUBTYPES) == Portal.LOOKUP_SUBTYPES

        results = []
        if not isinstance(portal_object, dict):
            return results
        if not (isinstance(portal_type, str) and portal_type):
            if isinstance(portal_type, dict):
                # It appears that the given portal_type is an actual schema dictionary.
                portal_type = self.schema_name(portal_type.get("title"))
            if not (isinstance(portal_type, str) and portal_type):
                if not (portal_type := self.get_schema_type(portal_object)):
                    return results
        if not callable(lookup_strategy):
            lookup_strategy = None if lookup_strategy is False else Portal._lookup_strategy
        for identifying_property in self.get_identifying_property_names(portal_type):
            if not (identifying_value := portal_object.get(identifying_property)):
                continue
            # The get_identifying_property_names call above ensures uuid is first if it is in the object.
            # And also note that ALL schemas do in fact have identifyingProperties which do in fact have
            # uuid, except for a couple "Test" ones, and (for some reason) SubmittedItem; otherwise we
            # might have a special case to check the Portal object explicitly for uuid, but no need.
            if identifying_property == "uuid":
                #
                # Note this idiosyncrasy with Portal paths: the only way we do NOT get a (HTTP 301) redirect
                # is if we use the lower-case-dashed-plural based version of the path, e.g. all of these:
                #
                # - /d13d06c1-218e-4f61-aaf0-91f226248b3c
                # - /d13d06c1-218e-4f61-aaf0-91f226248b3c/
                # - /FileFormat/d13d06c1-218e-4f61-aaf0-91f226248b3c
                # - /FileFormat/d13d06c1-218e-4f61-aaf0-91f226248b3c/
                # - /files-formats/d13d06c1-218e-4f61-aaf0-91f226248b3c
                #
                # Will result in a (HTTP 301) redirect to:
                #
                # - /files-formats/d13d06c1-218e-4f61-aaf0-91f226248b3c/
                #
                # Unfortunately, this code here has no reasonable way of getting that lower-case-dashed-plural
                # based name (e.g. file-formats) from the schema/portal type name (e.g. FileFormat); as the
                # information is contained, for this example, in the snovault.collection decorator for the
                # endpoint definition in smaht-portal/.../types/file_format.py. Unfortunately merely because
                # behind-the-scenes an extra round-trip HTTP request will occur, but happens automatically.
                # And note the disction of just using /{uuid} here rather than /{type}/{uuid} as in the else
                # statement below is not really necessary; just here for emphasis that this is all that's needed.
                #
                # TODO
                # Consider (from PR-308) writing a portal API for retrieving possible path formats.
                #
                if first_only is True:
                    results.append(f"/{portal_type}/{identifying_value}")
                else:
                    results.append(f"/{identifying_value}")
            elif isinstance(identifying_value, list):
                for identifying_value_item in identifying_value:
                    if identifying_value_item:
                        results.append(f"/{portal_type}/{identifying_value_item}")
            else:
                lookup_options = Portal.LOOKUP_UNDEFINED
                if schema := self.get_schema(portal_type):
                    if callable(lookup_strategy):
                        lookup_options, validator = lookup_strategy(self, portal_type, schema, identifying_value)
                        if callable(validator):
                            if validator(schema, identifying_property, identifying_value) is False:
                                continue
                    if pattern := schema.get("properties", {}).get(identifying_property, {}).get("pattern"):
                        if not re.match(pattern, identifying_value):
                            # If this identifying value is for a (identifying) property which has a
                            # pattern, and the value does NOT match the pattern, then do NOT include
                            # this value as an identifying path, since it cannot possibly be found.
                            continue
                if lookup_options == Portal.LOOKUP_UNDEFINED:
                    lookup_options = Portal.LOOKUP_DEFAULT
                if is_lookup_root_first(lookup_options):
                    results.append(f"/{identifying_value}")
                if is_lookup_specified_type(lookup_options) and portal_type:
                    results.append(f"/{portal_type}/{identifying_value}")
                if is_lookup_root(lookup_options) and not is_lookup_root_first(lookup_options):
                    results.append(f"/{identifying_value}")
                if is_lookup_subtypes(lookup_options):
                    for subtype_name in self.get_schema_subtype_names(portal_type):
                        results.append(f"/{subtype_name}/{identifying_value}")
            if (first_only is True) and results:
                return results
        return results

    @function_cache(maxsize=100, serialize_key=True)
    def get_identifying_path(self, portal_object: dict, portal_type: Optional[Union[str, dict]] = None,
                             lookup_strategy: Optional[Union[Callable, bool]] = None) -> Optional[str]:
        if identifying_paths := self.get_identifying_paths(portal_object, portal_type, first_only=True,
                                                           lookup_strategy=lookup_strategy):
            return identifying_paths[0]
        return None

    @function_cache(maxsize=100, serialize_key=True)
    def get_identifying_property_names(self, schema: Union[str, dict],
                                       portal_object: Optional[dict] = None) -> List[str]:
        """
        Returns the list of identifying property names for the given Portal schema, which may be
        either a schema name or a schema object. If a Portal object is also given then restricts this
        set of identifying properties to those which actually have values within this Portal object.
        Favors the uuid and identifier property names and defavors the aliases property name; no other
        ordering imposed. Returns empty list if no identifying properties or otherwise not found.
        """
        results = []
        if isinstance(schema, str):
            if not (schema := self.get_schema(schema)):
                return results
        elif not isinstance(schema, dict):
            return results
        if not (identifying_properties := get_identifying_properties(schema)):
            return results
        identifying_properties = list(set(identifying_properties))  # paranoid dedup
        identifying_properties = [*identifying_properties]  # copy so as not to change schema if given
        favored_identifying_properties = ["uuid", "identifier"]
        defavored_identifying_properties = ["aliases"]
        for favored_identifying_property in reversed(favored_identifying_properties):
            if favored_identifying_property in identifying_properties:
                identifying_properties.remove(favored_identifying_property)
                identifying_properties.insert(0, favored_identifying_property)
        for defavored_identifying_property in defavored_identifying_properties:
            if defavored_identifying_property in identifying_properties:
                identifying_properties.remove(defavored_identifying_property)
                identifying_properties.append(defavored_identifying_property)
        if isinstance(portal_object, dict):
            for identifying_property in [*identifying_properties]:
                if portal_object.get(identifying_property) is None:
                    identifying_properties.remove(identifying_property)
        return identifying_properties

    @staticmethod
    def _lookup_strategy(portal: Portal, type_name: str, schema: dict, value: str) -> (int, Optional[str]):
        #
        # Note this slightly odd situation WRT object lookups by submitted_id and accession:
        # -----------------------------+-----------------------------------------------+---------------+
        # PATH                         | EXAMPLE                                       | LOOKUP RESULT |
        # -----------------------------+-----------------------------------------------+---------------+
        # /submitted_id                | //UW_FILE-SET_COLO-829BL_HI-C_1               | NOT FOUND     |
        # /UnalignedReads/submitted_id | /UnalignedReads/UW_FILE-SET_COLO-829BL_HI-C_1 | FOUND         |
        # /SubmittedFile/submitted_id  | /SubmittedFile/UW_FILE-SET_COLO-829BL_HI-C_1  | FOUND         |
        # /File/submitted_id           | /File/UW_FILE-SET_COLO-829BL_HI-C_1           | NOT FOUND     |
        # -----------------------------+-----------------------------------------------+---------------+
        # /accession                   | /SMAFSFXF1RO4                                 | FOUND         |
        # /UnalignedReads/accession    | /UnalignedReads/SMAFSFXF1RO4                  | NOT FOUND     |
        # /SubmittedFile/accession     | /SubmittedFile/SMAFSFXF1RO4                   | NOT FOUND     |
        # /File/accession              | /File/SMAFSFXF1RO4                            | FOUND         |
        # -----------------------------+-----------------------------------------------+---------------+
        #
        def ref_validator(schema: Optional[dict],
                          property_name: Optional[str], property_value: Optional[str]) -> Optional[bool]:
            """
            Returns False iff objects of type represented by the given schema, CANNOT be referenced with
            a Portal path using the given property name and its given property value, otherwise returns None.

            For example, if the schema is for UnalignedReads and the property name is accession, then we will
            return False iff the given property value is NOT a properly formatted accession ID; otherwise, we
            will return None, which indicates that the caller (e.g. dcicutils.structured_data.Portal.ref_exists)
            will continue executing its default behavior, which is to check other ways in which the given type
            CANNOT be referenced by the given value, i.e. it checks other identifying properties for the type
            and makes sure any patterns (e.g. for submitted_id or uuid) are ahered to.

            The goal (in structured_data) being to detect if a type is being referenced in such a way that
            CANNOT possibly be allowed, i.e. because none of its identifying types are in the required form,
            if indeed there any requirements. It is assumed/guaranteed the given property name is indeed an
            identifying property for the given type.
            """
            if property_format := schema.get("properties", {}).get(property_name, {}).get("format"):
                if (property_format == "accession") and (property_name == "accession"):
                    if not Portal._is_accession_id(property_value):
                        return False
            return None

        DEFAULT_RESULT = (Portal.LOOKUP_DEFAULT, ref_validator)
        if not value:
            return DEFAULT_RESULT
        if not schema:
            if not isinstance(portal, Portal) or not (schema := portal.get_schema(type_name)):
                return DEFAULT_RESULT
        if schema_properties := schema.get("properties"):
            if schema_properties.get("accession") and Portal._is_accession_id(value):
                # Case: lookup by accession (only by root).
                return (Portal.LOOKUP_ROOT, ref_validator)
            elif schema_property_info_submitted_id := schema_properties.get("submitted_id"):
                if schema_property_pattern_submitted_id := schema_property_info_submitted_id.get("pattern"):
                    if re.match(schema_property_pattern_submitted_id, value):
                        # Case: lookup by submitted_id (only by specified type).
                        return (Portal.LOOKUP_SPECIFIED_TYPE, ref_validator)
        return DEFAULT_RESULT

    @staticmethod
    def _is_accession_id(value: str) -> bool:
        # This is here for now because of problems with circular dependencies.
        # See: smaht-portal/.../schema_formats.py/is_accession(instance) ...
        return isinstance(value, str) and re.match(r"^SMA[1-9A-Z]{9}$", value) is not None

    def url(self, url: str, raw: bool = False, database: bool = False) -> str:
        if not isinstance(url, str) or not url:
            return "/"
        elif (lowercase_url := url.lower()).startswith("http://") or lowercase_url.startswith("https://"):
            return url
        elif not (url := re.sub(r"/+", "/", url)).startswith("/"):
            url = "/"
        url = self.server + url if self.server else url
        if isinstance(raw, bool) and raw:
            url += ("&" if "?" in url else "?") + "frame=raw"
        if isinstance(database, bool) and database:
            url += ("&" if "?" in url else "?") + "datastore=database"
        return url

    def _kwargs(self, **kwargs) -> dict:
        if "headers" in kwargs:
            result_kwargs = {"headers": kwargs["headers"]}
        else:
            result_kwargs = {"headers": {"Content-type": Portal.MIME_TYPE_JSON, "Accept": Portal.MIME_TYPE_JSON}}
        if self.key_pair:
            result_kwargs["auth"] = self.key_pair
        if isinstance(timeout := kwargs.get("timeout"), int):
            result_kwargs["timeout"] = timeout
        return result_kwargs

    @staticmethod
    def _default_keys_file(app: Optional[str], env: Optional[str], server: Optional[str]) -> Optional[str]:
        def infer_app_from_env(env: str) -> Optional[str]:  # noqa
            if isinstance(env, str) and (lowercase_env := env.lower()):
                if app := [app for app in ORCHESTRATED_APPS if lowercase_env.startswith(app.lower())]:
                    return Portal._valid_app(app[0])
        if (app := Portal._valid_app(app)) or (app := infer_app_from_env(env)):
            keys_file = os.path.expanduser(os.path.join(Portal.KEYS_FILE_DIRECTORY, f".{app.lower()}-keys.json"))
            return keys_file if os.path.exists(keys_file) else None
        if not app:
            for app in ORCHESTRATED_APPS:
                if keys_file := Portal._default_keys_file(app, env, server):
                    if Portal._lookup_in_keys_file(keys_file, env, server)[0]:
                        return keys_file

    @staticmethod
    def _lookup_in_keys_file(keys_file: str, env: Optional[str], server: Optional[str],
                             raise_exception: bool = False) -> Tuple[Optional[dict], Optional[str]]:
        try:
            with io.open(keys_file := os.path.expanduser(keys_file)) as f:
                keys = json.load(f)
        except Exception:
            if raise_exception:
                raise Exception(f"Portal initialization error; cannot open keys-file: {keys_file}")
            return None, None
        if isinstance(env, str) and env and isinstance(key := keys.get(env), dict):
            return key, env
        elif (isinstance(server, str) and (server := Portal._normalize_server(server)) and
              (key := [keys[k] for k in keys if Portal._normalize_server(keys[k].get("server")) == server])):
            return key[0], env
        elif not env and len(keys) == 1 and (env := next(iter(keys))) and isinstance(key := keys[env], dict):
            return key, env
        else:
            if raise_exception:
                raise Exception(f"Portal initialization error;"
                                f" {env or server or None} not found in keys-file: {keys_file}")
            return None, None

    @staticmethod
    def _normalize_server(server: str) -> Optional[str]:
        prefix = ""
        if (lowercase_server := server.lower()).startswith("http://"):
            prefix = "http://"
        elif lowercase_server.startswith("https://"):
            prefix = "https://"
        if prefix:
            if (server := re.sub(r"/+", "/", server[len(prefix):])).startswith("/"):
                server = server[1:]
            if len(server) > 1 and server.endswith("/"):
                server = server[:-1]
            return prefix + server if server else None

    @staticmethod
    def _valid_app(app: Optional[str]) -> Optional[str]:
        if isinstance(app, str) and (app_lowercase := app.lower()):
            for value in ORCHESTRATED_APPS:
                if value.lower() == app_lowercase:
                    return value

    def _response(self, response: TestResponse) -> TestResponse:
        if response and isinstance(getattr(response.__class__, "json"), property):
            class TestResponseWrapper(TestResponse):  # For consistency change json property to method.
                def __init__(self, response, **kwargs):
                    super().__init__(**kwargs)
                    self._response = response
                def __getattr__(self, attr):  # noqa
                    return getattr(self._response, attr)
                def json(self):  # noqa
                    return self._response.json
                def raise_for_status(self):  # noqa
                    if self.status_code < 200 or self.status_code > 399:
                        raise requests.exceptions.HTTPError(f"HTTP Error: {self.status_code}", response=self)
            response = TestResponseWrapper(response)
        return response

    @staticmethod
    def _create_vapp(arg: Union[TestApp, VirtualApp, PyramidRouter, str] = None) -> TestApp:
        if isinstance(arg, TestApp):
            return arg
        elif isinstance(arg, VirtualApp):
            if not isinstance(arg.wrapped_app, TestApp):
                raise Exception("Portal._create_vapp VirtualApp argument error.")
            return arg.wrapped_app
        if isinstance(arg, PyramidRouter):
            router = arg
        elif isinstance(arg, str) or not arg:
            router = pyramid_get_app(arg or "development.ini", "app")
        else:
            raise Exception("Portal._create_vapp argument error.")
        return TestApp(router, {"HTTP_ACCEPT": Portal.MIME_TYPE_JSON, "REMOTE_USER": "TEST"})

    @staticmethod
    def create_for_testing(arg: Optional[Union[str, bool, List[dict], dict, Callable]] = None) -> Portal:
        if isinstance(arg, list) or isinstance(arg, dict) or isinstance(arg, Callable):
            return Portal(Portal._create_router_for_testing(arg))
        elif isinstance(arg, str) and arg.endswith(".ini"):
            return Portal(arg)
        elif arg == "local" or arg is True:
            minimal_ini_for_testing = "\n".join([
                "[app:app]\nuse = egg:encoded\nfile_upload_bucket = dummy",
                "sqlalchemy.url = postgresql://postgres@localhost:5441/postgres?host=/tmp/snovault/pgdata",
                "multiauth.groupfinder = encoded.authorization.smaht_groupfinder",
                "multiauth.policies = auth0 session remoteuser accesskey",
                "multiauth.policy.session.namespace = mailto",
                "multiauth.policy.session.use = encoded.authentication.NamespacedAuthenticationPolicy",
                "multiauth.policy.session.base = pyramid.authentication.SessionAuthenticationPolicy",
                "multiauth.policy.remoteuser.namespace = remoteuser",
                "multiauth.policy.remoteuser.use = encoded.authentication.NamespacedAuthenticationPolicy",
                "multiauth.policy.remoteuser.base = pyramid.authentication.RemoteUserAuthenticationPolicy",
                "multiauth.policy.accesskey.namespace = accesskey",
                "multiauth.policy.accesskey.use = encoded.authentication.NamespacedAuthenticationPolicy",
                "multiauth.policy.accesskey.base = encoded.authentication.BasicAuthAuthenticationPolicy",
                "multiauth.policy.accesskey.check = encoded.authentication.basic_auth_check",
                "multiauth.policy.auth0.use = encoded.authentication.NamespacedAuthenticationPolicy",
                "multiauth.policy.auth0.namespace = auth0",
                "multiauth.policy.auth0.base = encoded.authentication.Auth0AuthenticationPolicy"
            ])
        else:
            minimal_ini_for_testing = "[app:app]\nuse = egg:encoded\nsqlalchemy.url = postgresql://dummy\n"
        with temporary_file(content=minimal_ini_for_testing, suffix=".ini") as ini_file:
            return Portal(ini_file)

    @staticmethod
    def _create_router_for_testing(endpoints: Optional[List[Dict[str, Union[str, Callable]]]] = None) -> PyramidRouter:
        if isinstance(endpoints, dict):
            endpoints = [endpoints]
        elif isinstance(endpoints, Callable):
            endpoints = [{"path": "/", "method": "GET", "function": endpoints}]
        if not isinstance(endpoints, list) or not endpoints:
            endpoints = [{"path": "/", "method": "GET", "function": lambda request: {"status": "OK"}}]
        with PyramidConfigurator() as config:
            nendpoints = 0
            for endpoint in endpoints:
                if (endpoint_path := endpoint.get("path")) and (endpoint_function := endpoint.get("function")):
                    endpoint_method = endpoint.get("method", "GET")
                    def endpoint_wrapper(request):  # noqa
                        response = endpoint_function(request)
                        return PyramidResponse(json.dumps(response),
                                               content_type=f"{Portal.MIME_TYPE_JSON}; charset=utf-8")
                    endpoint_id = str(uuid())
                    config.add_route(endpoint_id, endpoint_path)
                    config.add_view(endpoint_wrapper, route_name=endpoint_id, request_method=endpoint_method)
                    nendpoints += 1
            if nendpoints == 0:
                return Portal._create_router_for_testing([])
            return config.make_wsgi_app()

    def start_for_testing(self, port: int = 7070, asynchronous: bool = False) -> Optional[Thread]:
        if isinstance(self.vapp, TestApp) and hasattr(self.vapp, "app") and isinstance(self.vapp.app, PyramidRouter):
            def start_server() -> None:  # noqa
                with wsgi_make_server("0.0.0.0", port or 7070, self.vapp.app) as server:
                    server.serve_forever()
            if asynchronous:
                server_thread = Thread(target=start_server)
                server_thread.daemon = True
                server_thread.start()
                return server_thread
            start_server()



import glob
import hashlib
import io
import os
import pathlib
from datetime import datetime
import random
import string
from tempfile import gettempdir as get_temporary_directory
from typing import List, Optional, Union
from uuid import uuid4 as uuid

HOME_DIRECTORY = str(pathlib.Path().home())


def search_for_file(file: str,
                    location: Union[str, pathlib.PosixPath, Optional[List[Union[str, pathlib.PosixPath]]]] = None,
                    recursive: bool = False,
                    single: bool = False,
                    order: bool = True) -> Union[List[str], Optional[str]]:
    """
    Searches for the existence of the given file name, first directly in the given directory or list
    of directories, if specified, and if not then just in the current (working) directory; if the
    given recursive flag is True then also searches all sub-directories of these directories;
    returns the full path name to the file if found. If the single flag is True then just the
    first file which is found is returns (as a string), or None if none; if the single flag
    is False, then all matched files are returned in a list, or and empty list if none.
    """
    def order_by_fewest_number_of_paths_and_then_alphabetically(paths: List[str]) -> List[str]:
        def order_by(path: str):
            return len(path.split(os.path.sep)), path
        return sorted(paths, key=order_by)

    if not (file and isinstance(file, (str, pathlib.PosixPath))):
        return None if single is True else []
    if os.path.isabs(file):
        if os.path.exists(file):
            return str(file) if single is True else [str(file)]
        return None if single is True else []
    files_found = []
    if not location:
        location = ["."]
    elif isinstance(location, (str, pathlib.PosixPath)):
        location = [location]
    elif not isinstance(location, list):
        location = []
    location_pruned = []
    for directory in location:
        if not isinstance(directory, str):
            if not isinstance(directory, pathlib.PosixPath):
                continue
            directory = str(directory)
        if not (directory := directory.strip()):
            continue
        if os.path.isfile(directory := os.path.abspath(os.path.normpath(directory))):
            # Actually, allow a file rather then a directory; assume its parent directory was intended.
            if not (directory := os.path.dirname(directory)):
                continue
        if directory not in location_pruned:
            location_pruned.append(directory)
    location = location_pruned
    for directory in location:
        if os.path.exists(os.path.join(directory, file)):
            file_found = os.path.abspath(os.path.normpath(os.path.join(directory, file)))
            if single is True:
                return file_found
            if file_found not in files_found:
                files_found.append(file_found)
    if recursive is True:
        for directory in location:
            if not directory.endswith("/**") and not file.startswith("**/"):
                path = f"{directory}/**/{file}"
            else:
                path = f"{directory}/{file}"
            files = glob.glob(path, recursive=True if recursive is True else False)
            if files:
                for file_found in files:
                    file_found = os.path.abspath(file_found)
                    if single is True:
                        return file_found
                    if file_found not in files_found:
                        files_found.append(file_found)
    if single is True:
        return files_found[0] if files_found else None
    elif order is True:
        return order_by_fewest_number_of_paths_and_then_alphabetically(files_found)
    else:
        return files_found


def normalize_path(value: Union[str, pathlib.Path], absolute: bool = False, expand_home: Optional[bool] = None) -> str:
    """
    Normalizes the given path value and returns the result; does things like remove redundant
    consecutive directory separators and redundant parent paths. If the given absolute argument
    is True than converts the path to an absolute path. If the given expand_home argument is False
    and if the path can reasonably be represented with a home directory indicator (i.e. "~"), then
    converts it to such. If the expand_home argument is True and path starts with the home directory
    indicator (i.e. "~") then expands it to the actual (absolute) home path of the caller. If the
    given path value is not actually even a string (or pathlib.Path) then returns an empty string.
    """
    if isinstance(value, pathlib.Path):
        value = str(value)
    elif not isinstance(value, str):
        return ""
    if not (value := value.strip()) or not (value := os.path.normpath(value)):
        return ""
    if expand_home is True:
        value = os.path.expanduser(value)
    elif (expand_home is False) and (os.name == "posix"):
        if value.startswith(home := HOME_DIRECTORY + os.sep):
            value = "~/" + value[len(home):]
        elif value == HOME_DIRECTORY:
            value = "~"
    if absolute is True:
        value = os.path.abspath(value)
    return value


def get_file_size(file: str, raise_exception: bool = True) -> Optional[int]:
    try:
        return os.path.getsize(file) if isinstance(file, str) else None
    except Exception:
        if raise_exception is True:
            raise
        return None


def get_file_modified_datetime(file: str, raise_exception: bool = True) -> Optional[datetime]:
    try:
        return datetime.fromtimestamp(os.path.getmtime(file)) if isinstance(file, str) else None
    except Exception:
        if raise_exception is True:
            raise
        return None


def are_files_equal(filea: str, fileb: str, raise_exception: bool = True) -> bool:
    """
    Returns True iff the contents of the two given files are exactly the same.
    """
    try:
        with open(filea, "rb") as fa:
            with open(fileb, "rb") as fb:
                chunk_size = 4096
                while True:
                    chunka = fa.read(chunk_size)
                    chunkb = fb.read(chunk_size)
                    if chunka != chunkb:
                        return False
                    if not chunka:
                        break
        return True
    except Exception:
        if raise_exception is True:
            raise
        return False


def compute_file_md5(file: str, raise_exception: bool = True) -> str:
    """
    Returns the md5 checksum for the given file.
    """
    if not isinstance(file, str):
        return ""
    try:
        md5 = hashlib.md5()
        with open(file, "rb") as file:
            for chunk in iter(lambda: file.read(4096), b""):
                md5.update(chunk)
        return md5.hexdigest()
    except Exception:
        if raise_exception is True:
            raise
        return ""


def compute_file_etag(file: str, raise_exception: bool = True) -> Optional[str]:
    """
    Returns the AWS S3 "etag" for the given file; this value is md5-like but
    not the same as a normal md5. We use this to compare that a file in S3
    appears to be the exact the same file as a local file.
    """
    try:
        with io.open(file, "rb") as f:
            return _compute_file_etag(f)
    except Exception:
        if raise_exception is True:
            raise
        return None


def _compute_file_etag(f: io.BufferedReader) -> str:
    # See: https://stackoverflow.com/questions/75723647/calculate-md5-from-aws-s3-etag
    MULTIPART_THRESHOLD = 8388608
    MULTIPART_CHUNKSIZE = 8388608
    # BUFFER_SIZE = 1048576
    # Verify some assumptions are correct
    # assert(MULTIPART_CHUNKSIZE >= MULTIPART_THRESHOLD)
    # assert((MULTIPART_THRESHOLD % BUFFER_SIZE) == 0)
    # assert((MULTIPART_CHUNKSIZE % BUFFER_SIZE) == 0)
    hash = hashlib.md5()
    read = 0
    chunks = None
    while True:
        # Read some from stdin, if we're at the end, stop reading
        bits = f.read(1048576)
        if len(bits) == 0:
            break
        read += len(bits)
        hash.update(bits)
        if chunks is None:
            # We're handling a multi-part upload, so switch to calculating
            # hashes of each chunk
            if read >= MULTIPART_THRESHOLD:
                chunks = b''
        if chunks is not None:
            if (read % MULTIPART_CHUNKSIZE) == 0:
                # Dont with a chunk, add it to the list of hashes to hash later
                chunks += hash.digest()
                hash = hashlib.md5()
    if chunks is None:
        # Normal upload, just output the MD5 hash
        etag = hash.hexdigest()
    else:
        # Multipart upload, need to output the hash of the hashes
        if (read % MULTIPART_CHUNKSIZE) != 0:
            # Add the last part if we have a partial chunk
            chunks += hash.digest()
        etag = hashlib.md5(chunks).hexdigest() + "-" + str(len(chunks) // 16)
    return etag


def create_random_file(file: Optional[str] = None, prefix: Optional[str] = None, suffix: Optional[str] = None,
                       nbytes: int = 1024, binary: bool = False, line_length: Optional[int] = None) -> str:
    """
    Write to the given file (name/path) some random content. If the given file is None then writes
    to a temporary file. In either case, returns the file written to. The of bytes written is 1024
    by default be can be specified with the nbytes argument; default to writing ASCII text but if
    the binary argument is True then writes binary data as well; if not binary the content is in
    lines of 80 characters each; use the line_length argumetn in this case to change the line length.
    """
    if not isinstance(nbytes, int) or nbytes < 0:
        nbytes = 0
    if not isinstance(file, str) or not file:
        if not isinstance(prefix, str):
            prefix = ""
        if not isinstance(suffix, str):
            suffix = ""
        file = f"{datetime.utcnow().strftime('%Y%m%d%H%M%S')}{str(uuid()).replace('-', '')}"
        file = os.path.join(get_temporary_directory(), file)
    with open(file, "wb" if binary is True else "w") as f:
        if binary is True:
            f.write(os.urandom(nbytes))
        else:
            if (not isinstance(line_length, int)) or (line_length < 1):
                line_length = 80
            line_length += 1
            nlines = nbytes // line_length
            nremainder = nbytes % line_length
            for n in range(nlines):
                f.write("".join(random.choices(string.ascii_letters + string.digits, k=line_length - 1)))
                f.write("\n")
            if nremainder > 1:
                f.write("".join(random.choices(string.ascii_letters + string.digits, k=nremainder - 1)))
            if nremainder > 0:
                f.write("\n")
    return file



from cryptography.hazmat.primitives import serialization
from datetime import datetime
import OpenSSL
import socket
import ssl
from typing import Optional, Tuple
from .misc_utils import future_datetime


def get_ssl_certificate_info(hostname: str,
                             raise_exception: bool = False,
                             test_mode_certificate_expiration_warning_days: int = 0) -> Optional[dict]:
    """
    Returns a dictionary containing various data points for the SSL certificate of the given
    hostname. If an error is encountered then returns None, or, if the given raise_exception
    argument is True then raises an exception. Returned dictionary contains (alphabetical order):

        active_at    issuer           owner_country
        exception    issuer_city      owner_entity
        expired      issuer_country   owner_state
        expires_at   issuer_entity    pem
        hostname     issuer_state     public_key_pem
        hostnames    owner            serial_number
        inactive     owner_city       invalid
    """
    hostname = _normalize_hostname(hostname)
    try:
        certificate_pem = _get_ssl_certificate_pem(hostname, raise_exception=False)
        certificate_info = _get_ssl_certificate_info_from_pem(
            certificate_pem,
            raise_exception=raise_exception,
            expires_soon_days=test_mode_certificate_expiration_warning_days
        )
        certificate_okay, certificate_exception = _is_ssl_certificate_okay(hostname, raise_exception=raise_exception)
        # The hostname from _get_ssl_certificate_info_from_pem is not necessarily exactly correct;
        # for example, for cgap-wolf.hms.harvard.edu it is imperva.com.
        certificate_info["hostname"] = hostname
        certificate_info["invalid"] = certificate_info["invalid"] or not certificate_okay
        if certificate_exception:
            certificate_info["exception"] = str(certificate_exception)
        return certificate_info
    except Exception as e:
        if raise_exception:
            raise e
        return None


_SSL_PORT = 443
_SSL_CERTIFICATE_EXPIRES_SOON_WARNING_DAYS = 7


def _get_ssl_certificate_pem(hostname: str, raise_exception: bool = False) -> Optional[str]:
    """
    Returns the SSL certificate as a PEM string for the given hostname. If an error is encountered
    then returns None, or, if the given raise_exception argument is True then raises an exception.
    """
    try:
        with ssl.create_connection((hostname, _SSL_PORT)) as socket_connection:
            # Note that we turn off certificate verification for getting the PEM here,
            # because we want the PEM regardless of whether or not the certificate is
            # valid; we do the real certificate sanity check in _is_ssl_certificate_okay.
            context = ssl.create_default_context()
            context.check_hostname = False
            context.verify_mode = ssl.CERT_NONE
            with context.wrap_socket(socket_connection, server_hostname=hostname) as socket_context:
                certificate = socket_context.getpeercert(binary_form=True)
                return ssl.DER_cert_to_PEM_cert(certificate)
    except Exception as e:
        if raise_exception:
            raise e
        return None


def _get_ssl_certificate_info_from_pem(pem_string: str,
                                       raise_exception: bool = False,
                                       expires_soon_days: int = 0) -> Optional[dict]:
    """
    Returns a dictionary containing various data points for the given SSL certificate string
    string in PEM format. If an error is encountered in parsing this given string then returns
    None by default, or raises and exception of the given raise_exception argument is True.
    """
    now = datetime.now()
    if expires_soon_days <= 0:
        expires_soon_days = _SSL_CERTIFICATE_EXPIRES_SOON_WARNING_DAYS

    def get_hostnames(certificate: OpenSSL.crypto.X509) -> list:
        """
        Returns the list of hostnames associated with the given SSL certificate. There can indeed
        be more than one hostname associated with a certificate, and we cannot determine the one
        primary one associated with it, so the caller (get_ssl_certificate_info) of this (outer)
        function (_get_ssl_certificate_info_from_pem) will set the hostname in the returned dictionary.
        """
        sans = []
        extension_count = certificate.get_extension_count()
        for i in range(extension_count):
            extension = certificate.get_extension(i)
            if extension.get_short_name() == b"subjectAltName":
                sans_list = str(extension).replace(" ", "").split(",")
                for san in sans_list:
                    if san.startswith("DNS:"):
                        sans.append(san[4:])
        return sorted(sans)

    try:
        certificate = OpenSSL.crypto.load_certificate(OpenSSL.crypto.FILETYPE_PEM, pem_string)

        subject = certificate.get_subject()
        common_name = subject.commonName
        hostnames = get_hostnames(certificate)
        owner_country = subject.countryName
        owner_state = subject.stateOrProvinceName
        owner_city = subject.localityName
        owner_entity = subject.organizationName
        owner = subject.organizationalUnitName or owner_entity
        if not owner:
            owner = common_name

        issuer = certificate.get_issuer()
        issuer_country = issuer.countryName
        issuer_state = issuer.stateOrProvinceName
        issuer_city = issuer.localityName
        issuer_entity = issuer.organizationName
        issuer = issuer.organizationalUnitName or issuer_entity

        not_before = certificate.get_notBefore().decode("UTF-8")
        not_after = certificate.get_notAfter().decode("UTF-8")
        active_at = datetime.strptime(not_before, "%Y%m%d%H%M%SZ")
        expires_at = datetime.strptime(not_after, "%Y%m%d%H%M%SZ")
        expires_soon = expires_at <= future_datetime(now=now, days=expires_soon_days)

        expired = now >= expires_at
        inactive = now <= active_at
        invalid = inactive or expired

        # Note we could not currently get detect specifically if the certificate is revoked;
        # could not get the recommeneded code working consistently; no great matter.

        serial_number = str(certificate.get_serial_number())
        public_key_pem = certificate.get_pubkey().to_cryptography_key().public_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PublicFormat.SubjectPublicKeyInfo
        ).decode("UTF-8")

        return {
            "name": None,  # Filled in by caller (get_ssl_certificate_info)
            "hostname": None,  # Filled in by caller (get_ssl_certificate_info)
            "hostnames": hostnames,
            "common_name": common_name,
            "owner": owner,
            "owner_entity": owner_entity,
            "owner_country": owner_country,
            "owner_state": owner_state,
            "owner_city": owner_city,
            "issuer": issuer,
            "issuer_entity": issuer_entity,
            "issuer_country": issuer_country,
            "issuer_state": issuer_state,
            "issuer_city": issuer_city,
            "active_at": active_at.strftime("%Y-%m-%d %H:%M:%S"),
            "expires_at": expires_at.strftime("%Y-%m-%d %H:%M:%S"),
            "pem": pem_string,
            "serial_number": serial_number,
            "public_key_pem": public_key_pem,
            "invalid": invalid,
            "inactive": inactive,
            "expired": expired,
            "expires_soon": expires_soon,
            "exception": None,
            "timestamp": now.strftime("%Y-%m-%d %H:%M:%S")
        }

    except Exception as e:
        if raise_exception:
            raise e
        return None


def _is_ssl_certificate_okay(hostname: str, raise_exception: bool = False) -> Tuple[bool, Optional[Exception]]:
    """
    Tries referencing the SSL certificate (in a couple different ways) of the given hostname
    so as to verify that the certificate is good, and if so, then returns a tuple containing,
    in order, True and None. If an error (exception) occurs then returns a tuple containing,
    in order, False and the Exception which was raised, or, if the given raise_exception
    argument is True then simply raises an exception.
    """
    try:
        # Note that with Python versions below 3.10 this fails on some sites (e.g. slashdot.org):
        # - ssl.get_server_certificate((hostname, _SSL_PORT))
        # https://stackoverflow.com/questions/70994539/ssl-errors-for-ssl-get-server-certificate-on-some-websites-but-not-on-others
        # FYI sample certificates: https://www.ssl.com/sample-valid-revoked-and-expired-ssl-tls-certificates/
        with socket.create_connection((hostname, _SSL_PORT)) as socket_connection:
            with ssl.create_default_context().wrap_socket(socket_connection, server_hostname=hostname):
                return True, None
    except Exception as e:
        if raise_exception:
            raise e
        return False, e


def _normalize_hostname(hostname: str) -> str:
    """
    If the given hostname actually looks like an URL then return just the hostname thereof,
    otherwise just return the given hostname; also lower-cases the value.
    """
    if hostname:
        hostname = hostname.lower()
        if hostname.startswith("https://"):
            hostname = hostname[8:]
        elif hostname.startswith("http://"):
            hostname = hostname[7:]
        slash = hostname.find("/")
        if slash > 0:
            hostname = hostname[:slash]
    return hostname



from dcicutils.misc_utils import normalize_spaces
from datetime import datetime, timedelta, timezone
from dateutil import parser as datetime_parser
from typing import Optional, Tuple, Union

TIMEZONE_LOCAL = datetime.now().astimezone().tzinfo  # type: datetime.timezone
TIMEZONE_LOCAL_NAME = TIMEZONE_LOCAL.tzname(None)  # type: str
TIMEZONE_LOCAL_OFFSET = TIMEZONE_LOCAL.utcoffset(None)  # type: datetime.timedelta
TIMEZONE_LOCAL_OFFSET_TOTAL_MINUTES = int(TIMEZONE_LOCAL_OFFSET.total_seconds()) // 60  # type: int
TIMEZONE_LOCAL_OFFSET_HOURS = TIMEZONE_LOCAL_OFFSET_TOTAL_MINUTES // 60  # type: int
TIMEZONE_LOCAL_OFFSET_MINUTES = TIMEZONE_LOCAL_OFFSET_TOTAL_MINUTES % 60  # type: int
TIMEZONE_LOCAL_SUFFIX = f"{TIMEZONE_LOCAL_OFFSET_HOURS:+03d}:{TIMEZONE_LOCAL_OFFSET_MINUTES:02d}"  # type: str

TIMEZONE_UTC = timezone.utc  # type: datetime.timezone
TIMEZONE_UTC_NAME = TIMEZONE_UTC.tzname(None)  # type: str
TIMEZONE_UTC_OFFSET = timedelta(0)  # type: datetime.timedelta
TIMEZONE_UTC_OFFSET_TOTAL_MINUTES = 0  # type: int
TIMEZONE_UTC_OFFSET_HOURS = 0  # type: int
TIMEZONE_UTC_OFFSET_MINUTES = 0  # type: int
TIMEZONE_UTC_SUFFIX = "Z"  # type: str


def parse_datetime_string(value: str) -> Optional[datetime]:
    """
    Parses the given string into a datetime object and returns it, or if ill-formated then returns None.
    The given string is assumed to be in the format "YYYY-MM-DD hh:mm:ss" and with an optional timezone
    suffix in format "+hh:mm" or "+hh". Also allowed is just a date of the format "YYYY-MM-DD" in which
    case a time of "00:00:00" is assumed. If no timezone is specified then the local timezone is assumed.
    """
    if not isinstance(value, str) or not (value := normalize_spaces(value)):
        return None
    tz_hours = -1
    tz_minutes = -1
    if value.rfind("T") > 0:
        value = value.replace("T", " ")
    if (space := value.find(" ")) > 0 and (value_suffix := value[space + 1:]):
        if (plus := value_suffix.rfind("+")) > 0 or (minus := value_suffix.rfind("-")) > 0:
            value = normalize_spaces(value[:space] + " " + value_suffix[:(plus if plus > 0 else minus)])
            if value_tz := normalize_spaces(value_suffix[(plus if plus > 0 else minus) + 1:]):
                if len(value_tz := value_tz.split(":")) == 2:
                    value_tz_hours = value_tz[0].strip()
                    value_tz_minutes = value_tz[1].strip()
                else:
                    value_tz_hours = value_tz[0].strip()
                    value_tz_minutes = "0"
                if value_tz_hours.isdigit() and value_tz_minutes.isdigit():
                    tz_hours = int(value_tz_hours)
                    tz_minutes = int(value_tz_minutes)
                    if not (plus > 0):
                        tz_hours = -tz_hours
    else:
        value = value + " 00:00:00"
    if tz_hours < 0 or tz_minutes < 0:
        tz_hours, tz_minutes = get_local_timezone_hours_minutes()
    try:
        try:
            dt = datetime.strptime(value, "%Y-%m-%d %H:%M:%S")
        except Exception:
            dt = datetime.strptime(value, "%Y-%m-%d %H:%M:%S.%f")
        tz = timezone(timedelta(hours=tz_hours, minutes=tz_minutes))
        return dt.replace(tzinfo=tz)
    except Exception:
        return None


def parse_date_string(value: str) -> Optional[datetime]:
    """
    Parses the given string into a datetime object representing only a date and
    returns it, or if ill-formated then returns None. The given string is assumed
    to be in the format "YYYY-MM-DD"; if a given string of this format is suffixed
    with a space or a "T" and ANYTHING else, then that trailing portion is ignored.
    """
    if isinstance(value, str) and (value := normalize_spaces(value)):
        if (separator := value.find(" ")) > 0 or (separator := value.find("T")) > 0:
            value = value[:separator]
        try:
            return datetime.strptime(value, "%Y-%m-%d")
        except Exception:
            pass


def normalize_datetime_string(value: str) -> Optional[str]:
    """
    Parses the given string into a datetime object and returns a string for that datetime in ISO-8601 format,
    or if ill-formated then returns None. The given string is assumed to be in the format "YYYY-MM-DD hh:mm:ss"
    and with an optional timezone suffix in format "+hh:mm" or "+hh". Also allowed is just a date of the
    format "YYYY-MM-DD" in which case a time of "00:00:00" is assumed. If no timezone is specified then
    the local timezone is assumed. The returned format looks like this: "2024-02-08T10:37:51-05:00"
    """
    dt = parse_datetime_string(value)
    return dt.isoformat() if dt else None


def normalize_date_string(value: str) -> Optional[str]:
    """
    Parses the given string into a datetime object representing only a date and returns a string for that
    date in ISO-8601 format, or if ill-formated then returns None. The given string is assumed to be in
    the format "YYYY-MM-DD"; but if a given string of this format is suffixed with a space followed by
    ANYTHING else, then that trailing portion is ignored. The returned format looks like this: "2024-02-08"
    """
    d = parse_date_string(value)
    return d.strftime("%Y-%m-%d") if d else None


def get_timezone(hours_or_timedelta: Union[int, timedelta], minutes: Optional[int] = None) -> timezone:
    try:
        if isinstance(hours_or_timedelta, timedelta):
            return timezone(hours_or_timedelta)
        return timezone(timedelta(hours=hours_or_timedelta, minutes=minutes or 0))
    except Exception:
        return TIMEZONE_LOCAL


def get_timezone_offset(tz: timezone) -> timedelta:
    try:
        return tz.utcoffset(None)
    except Exception:
        return TIMEZONE_LOCAL_OFFSET


def get_timezone_hours_minutes(tz: timezone) -> Tuple[int, int]:
    """
    Returns a tuple with the integer hours and minutes offset for the given timezone.
    If negative then only the hours is negative; the mintutes is always positive;
    this is okay because there are no timezones less than one hour from UTC.
    """
    tz_offset = get_timezone_offset(tz)
    tz_offset_total_minutes = int(tz_offset.total_seconds()) // 60
    tz_offset_hours = tz_offset_total_minutes // 60
    tz_offset_minutes = abs(tz_offset_total_minutes % 60)
    return tz_offset_hours, tz_offset_minutes


def get_utc_timezone() -> timezone:
    return TIMEZONE_UTC


def get_local_timezone() -> timezone:
    """
    Returns current/local timezone as a datetime.timezone object.
    """
    return TIMEZONE_LOCAL


def get_local_timezone_string() -> str:
    """
    Returns current/local timezone in format like: "-05:00".
    """
    return TIMEZONE_LOCAL_SUFFIX


def get_local_timezone_hours_minutes() -> Tuple[int, int]:
    """
    Returns a tuple with the integer hours and minutes offset for the current/local timezone.
    If negative then only the hours is negative; the mintutes is always positive;
    this is okay because there are no timezones less than one hour from UTC.
    """
    return TIMEZONE_LOCAL_OFFSET_HOURS, TIMEZONE_LOCAL_OFFSET_MINUTES


def parse_datetime(value: str, utc: bool = False, tz: Optional[timezone] = None) -> Optional[datetime]:
    """
    Parses the given string into a datetime, if possible, and returns that value,
    or None if not able to parse. The timezone of the returned datetime will be the
    local timezone; or if the given utc argument is True then it will be UTC; or if the
    given tz argument is a datetime.timezone then return datetime will be in that timezone.
    """
    if isinstance(value, datetime):
        return value
    elif not isinstance(value, str):
        return None
    try:
        # This dateutil.parser handles quite a wide variety of formats and suits our needs.
        value = datetime_parser.parse(value)
        if utc is True:
            # If the given utc argument is True then it trumps any tz argument if given.
            tz = timezone.utc
        if value.tzinfo is not None:
            # The given value had an explicit timezone specified.
            if isinstance(tz, timezone):
                return value.astimezone(tz)
            return value
        return value.replace(tzinfo=tz if isinstance(tz, timezone) else get_local_timezone())
    except Exception:
        return None


def format_datetime(value: datetime,
                    utc: bool = False,
                    tz: Optional[Union[timezone, bool]] = None,
                    iso: bool = False,
                    notz: bool = False,
                    noseconds: bool = False,
                    ms: bool = False,
                    verbose: bool = False,
                    noseparator: bool = False,
                    noday: bool = False,
                    nodate: bool = False,
                    notime: bool = False) -> str:
    """
    Returns the given datetime as a string in "YYYY:MM:DD hh:mm:ss tz" format, for
    example "2024-04-17 15:42:26 EDT". If the given notz argument is True then omits
    the timezone; if the noseconds argument is given the omits the seconds. If the given
    verbose argument is True then returns a really verbose version of the datetime, for
    example "Wednesday, April 17, 2024 | 15:42:26 EDT"; if the noseparator argument is
    True then omits the "|" separator; if the noday argument is True then omits the day
    of week part. The timezone of the returned datetime string will default to the local
    one; if the given utc argument is True then it will be UTC; or if the given tz
    argument is a datetime.timezone it will be in that timezone.
    """
    if nodate is True and notime is True:
        return ""
    if not isinstance(value, datetime):
        if not isinstance(value, str) or not (value := parse_datetime(value)):
            return ""
    try:
        if utc is True:
            tz = timezone.utc
        elif not isinstance(tz, timezone):
            if tz is True:
                notz = False
                tz = get_local_timezone()
            elif tz is False:
                notz = True
            tz = get_local_timezone()
        if noseconds is True:
            ms = False
        value = value.astimezone(tz)
        if iso:
            if notz is True:
                value = value.replace(tzinfo=None)
            if not (ms is True):
                value = value.replace(microsecond=0)
            if noseconds is True:
                if notz is True:
                    if nodate is True:
                        return value.strftime(f"%H:%M")
                    elif notime is True:
                        return value.strftime(f"%Y-%m-%d")
                    else:
                        return value.strftime(f"%Y-%m-%dT%H:%M")
                if len(tz := value.strftime("%z")) > 3:
                    tz = tz[:3] + ":" + tz[3:]
                if nodate is True:
                    return value.strftime(f"%H:%M") + tz
                elif notime is True:
                    return value.strftime(f"%Y-%m-%d") + tz
                else:
                    return value.strftime(f"%Y-%m-%dT%H:%M") + tz
            if nodate is True:
                if (not (notz is True)) and len(tz := value.strftime("%z")) > 3:
                    tz = tz[:3] + ":" + tz[3:]
                else:
                    tz = ""
                return value.strftime(f"%H:%M:%S{f'.%f' if ms is True else ''}") + tz
            elif notime is True:
                return value.strftime(f"%Y-%m-%d")
            else:
                return value.isoformat()
        if verbose:
            if nodate is True:
                return value.strftime(
                    f"%-I:%M{'' if noseconds is True else ':%S'}"
                    f"{f'.%f' if ms is True else ''} %p{'' if notz is True else ' %Z'}")
            elif notime is True:
                return value.strftime(f"{'' if noday is True else '%A, '}%B %-d, %Y")
            else:
                return value.strftime(
                    f"{'' if noday is True else '%A, '}%B %-d, %Y{'' if noseparator is True else ' |'}"
                    f" %-I:%M{'' if noseconds is True else ':%S'}"
                    f"{f'.%f' if ms is True else ''} %p{'' if notz is True else ' %Z'}")
        else:
            if nodate is True:
                return value.strftime(
                    f"%H:%M{'' if noseconds is True else ':%S'}"
                    f"{f'.%f' if ms is True else ''}{'' if notz is True else ' %Z'}")
            elif notime is True:
                return value.strftime(f"%Y-%m-%d")
            else:
                return value.strftime(
                    f"%Y-%m-%d %H:%M{'' if noseconds is True else ':%S'}"
                    f"{f'.%f' if ms is True else ''}{'' if notz is True else ' %Z'}")
    except Exception:
        return None


def format_date(value: datetime,
                utc: bool = False,
                tz: Optional[Union[timezone, bool]] = None,
                verbose: bool = False,
                noday: bool = False) -> str:
    return format_datetime(value, utc=utc, tz=tz, verbose=verbose, noday=noday, notime=True)


def format_time(value: datetime,
                utc: bool = False,
                iso: bool = False,
                tz: Optional[Union[timezone, bool]] = None,
                ms: bool = False,
                notz: bool = False,
                noseconds: bool = False,
                verbose: bool = False,
                noday: bool = False) -> str:
    return format_datetime(value, utc=utc, tz=tz, iso=iso, ms=ms, notz=notz,
                           noseconds=noseconds, verbose=verbose, nodate=True)



# Script to publish the Python package in the CURRENT git repo to PyPi.
# Does the following checks before allowing poetry to publish:
#
# 0. Current directory MUST be a git repo.
# 1. The git repo MUST NOT contain unstaged changes.
# 2. The git repo MUST NOT contain staged but uncommitted changes.
# 3. The git repo MUST NOT contain committed but unpushed changes.
# 4. The git repo package directories MUST NOT contain untracked files,
#    OR if they do contain untracked files then you must confirm this is OK.
# 5. The version being published must NOT have already been published.
#
# ASSUMES you have these credentials environment variables correctly set for PyPi publishing;
# although a --username and --password are also supported to set these via command-line.
#
# - PYPI_USER
# - PYPI_PASSWORD
#
# Prints a warning if the username is NOT "__token__" meaning a PyPi API token is being
# used, as using a simple username/password (as opposed to API token) is deprecated.
#
# Prompts for yes or no before publish is actually done. There is a --noconfirm
# option to skip this confimation, however it is only allowed when running in the
# context of GitHub actions - it checks for the GITHUB_ACTIONS environment variable.
#
# Prints warning if PY
#
# FYI: This was created late April 2023 after a junk file containing development
# logging output containing passwords was accidentally published to PyPi;
# item #4 above specifically addresses/prevents this. Perhaps better
# would be if publishing only happened via GitHub actions.

import argparse
import os
import requests
import subprocess
import sys
import toml

from typing import Tuple, Union


PYPI_BASE_URL = "https://pypi.org"
PYPI_API_TOKEN_USERNAME = "__token__"
DEBUG = False


def main() -> None:

    def is_github_actions_context():
        return "GITHUB_ACTIONS" in os.environ

    argp = argparse.ArgumentParser()
    argp.add_argument("--noconfirm", required=False, dest="noconfirm", action="store_true")
    argp.add_argument("--debug", required=False, dest="debug", action="store_true")
    argp.add_argument("--username", required=False, dest="username")
    argp.add_argument("--password", required=False, dest="password")
    argp.add_argument("--force-allow-username", required=False, action="store_true")
    args = argp.parse_args()

    if args.debug:
        global DEBUG
        DEBUG = True

    if args.noconfirm and not is_github_actions_context():
        ERROR_PRINT("The --noconfirm flag is only allowed within GitHub actions!")
        exit_with_no_action()

    if not verify_git_repo():
        exit_with_no_action()

    if not verify_unstaged_changes():
        exit_with_no_action()

    if not verify_uncommitted_changes():
        exit_with_no_action()

    if not verify_unpushed_changes():
        exit_with_no_action()

    if not verify_tagged():
        exit_with_no_action()

    if not verify_untracked_files():
        exit_with_no_action()

    package_name = get_package_name()
    package_version = get_package_version()

    if not verify_not_already_published(package_name, package_version):
        exit_with_no_action()

    if not args.noconfirm:
        if not answered_yes_to_confirmation(f"Do you want to publish {package_name} {package_version} to PyPi?"):
            exit_with_no_action()

    PRINT(f"Publishing {package_name} {package_version} to PyPi ...")

    if not publish_package(args.username, args.password, args.force_allow_username):
        exit_with_no_action()

    PRINT(f"Publishing {package_name} {package_version} to PyPi complete.")


def publish_package(pypi_username: str = None, pypi_password: str = None, force_allow_username: bool = False) -> bool:
    if not pypi_username:
        pypi_username = os.environ.get("PYPI_USER")
    if not pypi_password:
        pypi_password = os.environ.get("PYPI_PASSWORD")
    if not pypi_username or not pypi_password:
        ERROR_PRINT(f"No PyPi credentials; you should set the PYPI_USER and PYPI_PASSWORD environment variables.")
        return False
    if pypi_username != PYPI_API_TOKEN_USERNAME:
        if not force_allow_username:
            # Just in case someone really really needs this we will allow for now: --force-allow-username
            ERROR_PRINT(f"Publishing with username/pasword is no longer allowed; must use API token instead.")
            return False
        WARNING_PRINT(f"Publishing with username/pasword is NOT recommmended; use API token instead;"
                      f"only allowing because you said: --force-allow-username")
    poetry_publish_command = [
        "poetry", "publish",
        "--no-interaction", "--build",
        f"--username={pypi_username}", f"--password={pypi_password}"
    ]
    poetry_publish_results, status_code = execute_command(poetry_publish_command)
    PRINT("\n".join(poetry_publish_results))
    if status_code != 0:
        # TODO: Maybe retry once or twice (with prompt) if (perhaps spurious) failure.
        ERROR_PRINT(f"Publish to PyPi failed!")
        return False
    return True


def verify_git_repo() -> bool:
    """
    If this (the current directory) looks like a git repo then return True,
    otherwise prints an error message and returns False.
    """
    _, status = execute_command("git rev-parse --is-inside-work-tree")
    if status != 0:
        ERROR_PRINT("You are not in a git repo directory!")
        return False
    return True


def verify_unstaged_changes() -> bool:
    """
    If the current git repo has NO unstaged changes then returns True,
    otherwise prints an error message and returns False. HOWEVER, we DO
    allow unstaged changes to just the file gitinfo.json if such exists; to
    allow GitHub Actions to update with latest git (repo, branch, commit) info.
    """
    git_diff_results, _ = execute_command("git diff --name-only")
    if git_diff_results and not (len(git_diff_results) == 1 and
                                 os.path.basename(git_diff_results[0]) == "gitinfo.json"):
        ERROR_PRINT("You have changes to this branch that you have not staged for commit.")
        return False
    return True


def verify_uncommitted_changes() -> bool:
    """
    If the current git repo has no staged but uncommitted changes then returns True,
    otherwise prints an error message and returns False.
    """
    git_diff_staged_results, _ = execute_command("git diff --staged")
    if git_diff_staged_results:
        ERROR_PRINT("You have changes to this branch that you have staged but not committed.")
        return False
    return True


def verify_unpushed_changes() -> bool:
    """
    If the current git repo committed but unpushed changes then returns True,
    otherwise prints an error message and returns False.
    """
    git_uno_results, _ = execute_command("git status -uno", lines_containing="is ahead of")
    if git_uno_results:
        ERROR_PRINT("You have committed changes to this branch that you committed but not pushed.")
        return False
    return True


def verify_tagged() -> bool:
    """
    If the current git repo has a tag as its most recent commit then returns True,
    otherwise prints an error message and returns False.
    """
    git_most_recent_commit, _ = execute_command("git log -1 --decorate", lines_containing="tag:")
    if not git_most_recent_commit:
        ERROR_PRINT("You can only publish a tagged commit.")
        return False
    return True


def verify_untracked_files() -> bool:
    """
    If the current git repo has no untracked files then returns True,
    otherwise prints an error message, with the list of untracked files,
    and prompts the user for a yes/no confirmation on whether or to
    continue, and returns True for a yes response, otherwise returns False.
    """
    untracked_files = get_untracked_files()
    if untracked_files:
        PRINT(f"You are about to PUBLISH the following ({len(untracked_files)})"
              f" UNTRACKED file{'' if len(untracked_files) == 1 else 's'} -> SECURITY risk:")
        for untracked_file in untracked_files:
            PRINT(f"-- {untracked_file}")
        PRINT("DO NOT continue UNLESS you KNOW what you are doing!")
        if not answered_yes_to_confirmation("Do you really want to continue?"):
            return False
    return True


def verify_not_already_published(package_name: str, package_version: str) -> bool:
    """
    If the given package and version has NOT already been published to PyPi then returns True,
    otherwise prints an error message and returns False.
    """
    url = f"{PYPI_BASE_URL}/pypi/{package_name}/json"
    DEBUG_PRINT(f"curl {url}")
    try:
        response = requests.get(url).json()["releases"]
    except Exception:
        WARNING_PRINT(f"Cannot determine if this package ({package_name} {package_version})"
                      f" has already been published or not; ASSUMING NOT.")
        return True
    if response.get(package_version):
        ERROR_PRINT(f"Package {package_name} {package_version} has already been published to PyPi.")
        return False
    return True


def verify_not_already_published_obsolete_no_longer_works_20250110(package_name: str, package_version: str) -> bool:
    """
    If the given package and version has not already been published to PyPi then returns True,
    otherwise prints an error message and returns False.
    """
    url = f"{PYPI_BASE_URL}/project/{package_name}/{package_version}/"
    DEBUG_PRINT(f"curl {url}")
    response = requests.get(url)
    if response.status_code == 200:
        ERROR_PRINT(f"Package {package_name} {package_version} has already been published to PyPi.")
        return False
    return True


def get_untracked_files() -> list:
    """
    Returns a list of untracked files for the current git repo; empty list if no untracked changes.
    We ignore __pycache__ directories for this which are already excluded by poetry publish.
    We ignore the .gitignore file for this.
    """
    package_directories = get_package_directories()
    untracked_files = []
    for package_directory in package_directories:
        # Note that the output of "git status -s --ignored" looks something like this:
        # ?? chalicelib_fourfront/some_untracked_file.py
        # !! chalicelib_fourfront/__pycache__/
        # Note that the --ignored option ignores the .gitignore file.
        git_status_results, _ = execute_command(f"git status -s --ignored {package_directory}")
        for git_status_result in git_status_results:
            if git_status_result and (git_status_result.startswith("??") or git_status_result.startswith("!!")):
                untracked_file = git_status_result[2:].strip()
                if untracked_file:
                    # Ignore any __pycache__ directories as they are already ignored by poetry publish.
                    if os.path.isdir(untracked_file) and os.path.basename(untracked_file.rstrip("/")) == "__pycache__":
                        continue
                    untracked_files.append(untracked_file)
    return untracked_files


def get_package_version() -> str:
    """
    Returns the tag name of the most recently created tag in the current git repo.
    """
    tag_commit, _ = execute_command("git rev-list --tags --max-count=1")
    tag_name, _ = execute_command(f"git  describe --tags  {tag_commit[0]}")
    package_version = tag_name[0]
    if package_version.startswith("v"):
        package_version = package_version[1:]
    return package_version


def get_package_name() -> str:
    """
    Returns the base name of the current git repo name.
    """
    # No, the package name should come from pyproject.toml not from the git repo name ...
    # package_name, _ = execute_command("git config --get remote.origin.url")
    # package_name = os.path.basename(package_name[0])
    # if package_name.endswith(".git"):
    #     package_name = package_name[:-4]
    # return package_name
    pyproject_toml = get_pyproject_toml()
    return pyproject_toml["tool"]["poetry"]["name"]


def get_package_directories() -> list:
    """
    Returns a list of directories constituting the Python package of the current repo,
    according to the pyproject.toml file.
    """
    package_directories = []
    pyproject_toml = get_pyproject_toml()
    pyproject_package_directories = pyproject_toml["tool"]["poetry"]["packages"]
    for pyproject_package_directory in pyproject_package_directories:
        package_directory = pyproject_package_directory.get("include")
        if package_directory:
            package_directory_from = pyproject_package_directory.get("from")
            if package_directory_from:
                package_directory = os.path.join(package_directory_from, package_directory)
            package_directories.append(package_directory)
    return package_directories


def get_pyproject_toml() -> list:
    with open("pyproject.toml", "r") as f:
        return toml.load(f)


def execute_command(command_argv: Union[list, str], lines_containing: str = None) -> Tuple[list, int]:
    """
    Executes the given command as a command-line subprocess, and returns a tuple whose first element
    is the list of lines from the output of the command, and the second element is the exit status code.
    """
    def cleanup_funny_output(output: str) -> str:
        return output.replace("('", "").replace("',)", "").replace("\\n\\n", "\n").replace("\\n", "\n")

    if isinstance(command_argv, str):
        command_argv = [arg for arg in command_argv.split(" ") if arg.strip()]
    DEBUG_PRINT(" ".join(command_argv))
    result = subprocess.run(command_argv, stdout=subprocess.PIPE,
                            stderr=subprocess.STDOUT)
    lines = result.stdout.decode("utf-8").split("\n")
    if lines_containing:
        lines = [line for line in lines if lines_containing in line]
    return [cleanup_funny_output(line.strip()) for line in lines if line.strip()], result.returncode


def answered_yes_to_confirmation(message: str) -> bool:
    """
    Prompts the user with the given message and asks for a yes or no answer,
    and if yes is the user response then returns True, otherwise returns False.
    """
    response = input(f"{message} [yes | no]: ").lower()
    if response == "yes":
        return True
    return False


def exit_with_no_action() -> None:
    """
    Exits this process immediately with status 1;
    first prints a message saying no action was taken.
    """
    PRINT("Exiting without taking action.")
    sys.exit(1)


PRINT = print


def WARNING_PRINT(s):
    PRINT(f"WARNING: {s}")


def ERROR_PRINT(s):
    PRINT(f"ERROR: {s}")


def DEBUG_PRINT(s):
    if DEBUG:
        PRINT(f"DEBUG: {s}")


if __name__ == "__main__":
    main()



# Though dcicutils is not dependent on numpy, elasticsearch pulls it in iff it is installed,
# and if it is numpy 2.x the numpy.float_ constant has been retired and any reference to it
# yields an error from numpy (AttributeError: np.float_ was removed in the NumPy 2.0 release.
# Use np.float64 instead); this reference to numpy.float_ occurs in elasticsearch/serializer.py.
# and we short-circuit it here by explicitly setting numpy.float_ to numpyh.float64.
try:
    import numpy
    numpy.float_ = numpy.float64
except Exception:
    pass
